"""
Services defined by the Workspaces system, to be used by our APIs and client programs.
"""
import stat
import subprocess
from tempfile import mkdtemp

from .interfaces import *
from .schema import *


class WorkflowService(WorkflowServiceIF):
    def execute(self, workflow_name: str, argument: Dict, files: List[Path]):
        """
        Here's the implementation from wksp0 that does this:

            # 1. look up the workflow info for this workflow name
            info = self.db.lookup_workflow_definition(workflow_name)

            # 2. create and save this request to the database
            record = WorkflowRequest(...)

            # 3. render the templates to files
            contents = info.render_templates(argument, files)

            # 4. serialize the templated files
            temp_folder = self._prepare_files_for_condor(contents)

            # 5. execute condor and get the log file
            log_file = self._execute_prepared(temp_folder)

            # 6. start reading the logs
            return HTCondorWorkflowEventStream(log_file)

        Of these, step 6 is going to have to be the most different, because
        that's launching the new CLI wf_monitor program that Nathan wrote.
        The other steps are more similar than different, though figuring out
        where to put the temp files will be interesting.
        """
        raise NotImplementedError

    @staticmethod
    def _prepare_files_for_condor(files: Dict[str, str]) -> Path:
        """
        Place the files for Condor into a new temp directory and returns the directory.

        :param files:  a dictionary of filename -> content
        :return:       a Path
        """
        # 1. create a temporary directory
        temp_folder = Path(mkdtemp(dir=str(Path.home() / "tmp")))

        # 2. spool each of the temp files to it
        for name, content in files.items():
            (temp_folder / name).write_text(content)

        # 3. make any scripts in there executable
        for file in temp_folder.glob('*.sh'):
            file.chmod(file.stat().st_mode | stat.S_IEXEC)

        # finished, return folder
        return temp_folder

    @staticmethod
    def _execute_prepared(folder: Path) -> Path:
        """
        Execute HTCondor using the named folder as the source of the files.

        :param folder:  the path to the folder to execute
        :return:        the path to the log file
        """
        print(f'executing on folder {folder}')

        # some file in here should end in .dag; that file is our dagman input
        dagman = list(folder.glob('*.dag'))[0]

        # ensure the log file exists
        logfile = folder / 'condor.log'
        logfile.touch()

        # submit
        subprocess.run(['condor_submit_dag', str(dagman)], cwd=str(folder.absolute()))

        # return the logfile
        return logfile

    # The next few things that need to happen here are in response to
    # WorkflowEvents that we receive from wf_monitor as the workflow
    # execution evolves. So we have to set up listening at some point
    # in this class
    def __init__(self):
        # 1. Start listening for events from the wf_monitor stream
        # self.channel = channels.workflow_events.listen(self.on_workflow_event)
        raise NotImplementedError

    def on_workflow_event(self, event: WorkflowEvent):
        # 1. log that we received this event, somehow
        # 2. update the WorkflowRequest record with the state we got
        # 3. do per-event-type stuff, such as level change events, database
        #    updates, and logging
        if event.type == WorkflowEventType.SUBMITTED:
            # this would be a good place to emit a level-changed event
            # for the capability system, such as:
            #
            #   channels.capability_events.send(...)
            #
            # however, for now we can pass
            pass
        elif event.type == WorkflowEventType.EXECUTING:
            # another good place to emit a level-changed event
            pass
        elif event.type == WorkflowEventType.TERMINATED:
            # another level-change event opportunity
            # also, we must now locate the temporary directory and
            # remove it
            pass
        # there is an event type of OTHER which we can basically
        # ignore for now

    def __exit__(self, exc_type, exc_val, exc_tb):
        # this method lets us use "with WorkflowService as ws" type syntax
        # we should close our AMQP channel gracefully here
        # self.channel.close()
        pass


class WorkflowInfo(WorkflowInfoIF):
    """
    Implements WorkflowInfo, backed by a relational database, using SQL Alchemy.
    """

    def __init__(self, session):
        self.session = session

    def lookup_workflow_definition(self, name: str) -> Workflow:
        return self.session.query(Workflow).get(name)

    def all_workflows(self) -> List[Workflow]:
        return self.session.query(Workflow).all()
